@@ -41,6 +41,8 @@
 #include "eim.h"
 #include "keyboard.h"
 
+#define MAX_LENGTH 40
+
 FCITX_EXPORT_API
 FcitxIMClass ime = {
     FcitxHangulCreate,
@@ -60,16 +62,23 @@ boolean LoadHangulConfig(FcitxHangulConfig* fs);
 static void SaveHangulConfig(FcitxHangulConfig* fs);
 static void ConfigHangul(FcitxHangul* hangul);
 static void FcitxHangulUpdatePreedit(FcitxHangul* hangul);
-static void FcitxHangulUpdateLookupTable(FcitxHangul* hangul);
+static void FcitxHangulCleanLookupTable(FcitxHangul* hangul);
+static void FcitxHangulUpdateLookupTable(FcitxHangul* hangul, boolean checkSurrounding);
 static void FcitxHangulFlush(FcitxHangul* hangul);
 static void FcitxHangulToggleHanja(void* arg);
 static boolean FcitxHangulGetHanja(void* arg);
 static void FcitxHangulResetEvent(void* arg);
 static char* FcitxHangulUcs4ToUtf8(FcitxHangul* hangul, const ucschar* ucsstr, int length);
 static void FcitxHangulUpdateHanjaStatus(FcitxHangul* hangul);
-static void FcitxHangulUpdateSurroundingText(void* arg);
 
-size_t ucs4_strlen(const ucschar* str)
+static inline void FcitxHangulFreeHanjaList(FcitxHangul* hangul) {
+    if (hangul->hanjaList) {
+        hanja_list_delete (hangul->hanjaList);
+        hangul->hanjaList = NULL;
+    }
+}
+
+static inline size_t ucs4_strlen(const ucschar* str)
 {
     size_t len = 0;
     while(*str) {
@@ -129,6 +138,9 @@ void FcitxHangulReset (void* arg)
     FcitxHangul* hangul = (FcitxHangul*) arg;
     ustring_clear(hangul->preedit);
     hangul_ic_reset(hangul->ic);
+    if (hangul->hanjaList) {
+        FcitxHangulCleanLookupTable(hangul);
+    }
 }
 
 /**
@@ -142,19 +154,30 @@ void FcitxHangulReset (void* arg)
 INPUT_RETURN_VALUE FcitxHangulDoInput(void* arg, FcitxKeySym sym, unsigned int state)
 {
     FcitxHangul* hangul = (FcitxHangul*) arg;
+    FcitxInstance* instance = hangul->owner;
 
     if (FcitxHotkeyIsHotKey(sym, state, hangul->fh.hkHanjaMode)) {
-        FcitxUIUpdateStatus(hangul->owner, "hanja");
-        if (hangul->fh.hkHanjaMode)
-            return IRV_DISPLAY_CANDWORDS;
-        else
-            return IRV_DO_NOTHING;
+        if (hangul->hanjaList == NULL) {
+            FcitxHangulUpdateLookupTable(hangul, true);
+        }
+        else {
+            FcitxHangulCleanLookupTable(hangul);
+        }
+        return IRV_DISPLAY_MESSAGE;
     }
 
     if (sym == FcitxKey_Shift_L || sym == FcitxKey_Shift_R)
         return IRV_TO_PROCESS;
 
-    int s = hangul->fh.hkHanjaMode[0].state | hangul->fh.hkHanjaMode[1].state;
+    FcitxGlobalConfig* config = FcitxInstanceGetGlobalConfig(hangul->owner);
+    const FcitxHotkey* prevPage = FcitxConfigPrevPageKey(instance, config);
+    const FcitxHotkey* nextPage = FcitxConfigNextPageKey(instance, config);
+    int s = hangul->fh.hkHanjaMode[0].state | hangul->fh.hkHanjaMode[1].state
+          | config->prevWord[0].state | config->prevWord[1].state
+          | config->nextWord[0].state | config->nextWord[1].state
+          | prevPage[0].state | prevPage[1].state
+          | nextPage[0].state | nextPage[1].state;
+
     if (s & FcitxKeyState_Ctrl) {
         if (sym == FcitxKey_Control_L || sym == FcitxKey_Control_R)
             return IRV_TO_PROCESS;
@@ -176,34 +199,89 @@ INPUT_RETURN_VALUE FcitxHangulDoInput(void* arg, FcitxKeySym sym, unsigned int s
             return IRV_TO_PROCESS;
     }
 
-    s = FcitxKeyState_Ctrl | FcitxKeyState_Alt | FcitxKeyState_Shift | FcitxKeyState_Super | FcitxKeyState_Hyper;
-    if (state & s) {
-        FcitxHangulFlush (hangul);
-        return IRV_TO_PROCESS;
-    }
-
     FcitxInputState* input = FcitxInstanceGetInputState(hangul->owner);
     FcitxCandidateWordList* candList = FcitxInputStateGetCandidateList(input);
     int candSize = FcitxCandidateWordGetListSize(candList);
 
     if (candSize > 0) {
-        FcitxGlobalConfig* config = FcitxInstanceGetGlobalConfig(hangul->owner);
+        if (FcitxHotkeyIsHotKey(sym, state, prevPage)) {
+            if (FcitxCandidateWordHasPrev(candList)) {
+                FcitxCandidateWordGetFocus(candList, true);
+            }
+            if (FcitxCandidateWordGoPrevPage(candList)) {
+                FcitxCandidateWordSetType(FcitxCandidateWordGetByIndex(candList, 0), MSG_CANDIATE_CURSOR);
+                return IRV_FLAG_UPDATE_INPUT_WINDOW;
+            } else {
+                return IRV_DO_NOTHING;
+            }
+        } else if (FcitxHotkeyIsHotKey(sym, state, nextPage)) {
+            if (FcitxCandidateWordHasNext(candList)) {
+                FcitxCandidateWordGetFocus(candList, true);
+            }
+            if (FcitxCandidateWordGoNextPage(candList)) {
+                FcitxCandidateWordSetType(FcitxCandidateWordGetByIndex(candList, 0), MSG_CANDIATE_CURSOR);
+                return IRV_FLAG_UPDATE_INPUT_WINDOW;
+            } else {
+                return IRV_DO_NOTHING;
+            }
+        }
 
-        if (FcitxHotkeyIsHotKey(sym, state, config->hkPrevPage) || FcitxHotkeyIsHotKey(sym, state, config->hkNextPage))
-            return IRV_TO_PROCESS;
+        FcitxCandidateWord* candWord = NULL;
+        if (FcitxHotkeyIsHotKey(sym, state, config->nextWord)) {
+            candWord = FcitxCandidateWordGetFocus(candList, true);
+            candWord = FcitxCandidateWordGetNext(candList, candWord);
+            if (!candWord) {
+                FcitxCandidateWordSetPage(candList, 0);
+                candWord = FcitxCandidateWordGetCurrentWindow(candList);
+            } else {
+                FcitxCandidateWordSetFocus(
+                    candList, FcitxCandidateWordGetIndex(candList,
+                                                        candWord));
+            }
+        } else if (FcitxHotkeyIsHotKey(sym, state, config->prevWord)) {
+            candWord = FcitxCandidateWordGetFocus(candList, true);
+            candWord = FcitxCandidateWordGetPrev(candList, candWord);
+            if (!candWord) {
+                candWord = FcitxCandidateWordGetLast(candList);
+            }
+            FcitxCandidateWordSetFocus(
+                candList, FcitxCandidateWordGetIndex(candList,
+                                                     candWord));
+        }
+        if (candWord) {
+            FcitxCandidateWordSetType(candWord, MSG_CANDIATE_CURSOR);
+            return IRV_FLAG_UPDATE_INPUT_WINDOW;
+        }
 
         if (FcitxHotkeyIsHotKeyDigit(sym, state))
             return IRV_TO_PROCESS;
 
         if (FcitxHotkeyIsHotKey(sym, state , FCITX_ENTER)) {
-            return FcitxCandidateWordChooseByIndex(FcitxInputStateGetCandidateList(input), 0);
+            do {
+                candWord = FcitxCandidateWordGetFocus(candList, true);
+                if (!candWord) {
+                    break;
+                }
+                // FcitxLog(INFO, "%d", FcitxCandidateWordGetIndex(candList, candWord));
+                return FcitxCandidateWordChooseByTotalIndex(candList,
+                                                            FcitxCandidateWordGetIndex(candList, candWord));
+            } while(0);
+            return FcitxCandidateWordChooseByIndex(candList, 0);
         }
 
-         if (!hangul->fh.hanjaMode) {
-            return IRV_DONOT_PROCESS;
+        if (!hangul->fh.hanjaMode) {
+            FcitxHangulCleanLookupTable(hangul);
         }
     }
 
+    s = FcitxKeyState_Ctrl | FcitxKeyState_Alt | FcitxKeyState_Shift | FcitxKeyState_Super | FcitxKeyState_Hyper;
+    if (state & s) {
+        FcitxHangulFlush (hangul);
+        FcitxHangulUpdatePreedit(hangul);
+        FcitxUIUpdateInputWindow(hangul->owner);
+        return IRV_TO_PROCESS;
+    }
+
     bool keyUsed = false;
     if (FcitxHotkeyIsHotKey(sym, state, FCITX_BACKSPACE)) {
         keyUsed = hangul_ic_backspace (hangul->ic);
@@ -215,6 +293,10 @@ INPUT_RETURN_VALUE FcitxHangulDoInput(void* arg, FcitxKeySym sym, unsigned int s
             }
         }
     } else {
+        if (ustring_length(hangul->preedit) >= MAX_LENGTH) {
+            FcitxHangulFlush(hangul);
+        }
+
         keyUsed = hangul_ic_process(hangul->ic, sym);
         boolean notFlush = false;
 
@@ -283,6 +365,7 @@ void FcitxHangulUpdatePreedit(FcitxHangul* hangul)
     FcitxMessages* preedit = FcitxInputStateGetPreedit(input);
     FcitxMessages* clientPreedit = FcitxInputStateGetClientPreedit(input);
     FcitxInstanceCleanInputWindowUp(hangul->owner);
+    FcitxInputStateSetShowCursor(input, true);
     const ucschar *hic_preedit = hangul_ic_get_preedit_string (hangul->ic);
 
     char* pre1 = FcitxHangulUcs4ToUtf8(hangul, ustring_begin(hangul->preedit), ustring_length(hangul->preedit));
@@ -290,20 +373,28 @@ void FcitxHangulUpdatePreedit(FcitxHangul* hangul)
     FcitxInputContext* ic = FcitxInstanceGetCurrentIC(hangul->owner);
     FcitxProfile* profile = FcitxInstanceGetProfile(hangul->owner);
 
+    size_t preeditLen = 0;
+
+    boolean clientPreeditNotAvail = (ic && ((ic->contextCaps & CAPACITY_PREEDIT) == 0 || !profile->bUsePreedit));
+
     if (pre1 && pre1[0] != 0) {
-        if (ic && ((ic->contextCaps & CAPACITY_PREEDIT) == 0 || !profile->bUsePreedit))
+        size_t len1 = strlen(pre1);
+        if (clientPreeditNotAvail)
             FcitxMessagesAddMessageAtLast(preedit, MSG_INPUT, "%s", pre1);
         FcitxMessagesAddMessageAtLast(clientPreedit, MSG_INPUT, "%s", pre1);
+        preeditLen += len1;
     }
 
     if (pre2 && pre2[0] != '\0') {
-        if (ic && ((ic->contextCaps & CAPACITY_PREEDIT) == 0 || !profile->bUsePreedit))
+        size_t len2 = strlen(pre2);
+        if (clientPreeditNotAvail)
             FcitxMessagesAddMessageAtLast(preedit, MSG_INPUT | MSG_HIGHLIGHT, "%s", pre2);
         FcitxMessagesAddMessageAtLast(clientPreedit, MSG_INPUT | MSG_HIGHLIGHT, "%s", pre2);
+        preeditLen += len2;
     }
 
-    FcitxInputStateSetCursorPos(input, 0);
-    FcitxInputStateSetClientCursorPos(input, 0);
+    FcitxInputStateSetCursorPos(input, clientPreeditNotAvail ? preeditLen : 0);
+    FcitxInputStateSetClientCursorPos(input, preeditLen);
 
     if (pre1)
         free(pre1);
@@ -314,7 +405,7 @@ void FcitxHangulUpdatePreedit(FcitxHangul* hangul)
 
 HanjaList* FcitxHangulLookupTable(FcitxHangul* hangul, const char* key, int method)
 {
-    HanjaList* list;
+    HanjaList* list = NULL;
 
     if (key == NULL)
         return NULL;
@@ -384,7 +475,13 @@ GetSubstring (const char* str, long p1, long p2)
     return substring;
 }
 
-void FcitxHangulUpdateLookupTable(FcitxHangul* hangul)
+void FcitxHangulCleanLookupTable(FcitxHangul* hangul) {
+    FcitxInstanceCleanInputWindowDown(hangul->owner);
+    // FcitxUIUpdateInputWindow(hangul->owner);
+    FcitxHangulFreeHanjaList(hangul);
+}
+
+void FcitxHangulUpdateLookupTable(FcitxHangul* hangul, boolean checkSurrounding)
 {
     char* surroundingStr = NULL;
     char* utf8;
@@ -395,10 +492,7 @@ void FcitxHangulUpdateLookupTable(FcitxHangul* hangul)
     unsigned int cursorPos;
     unsigned int anchorPos;
 
-    if (hangul->hanjaList != NULL) {
-        hanja_list_delete (hangul->hanjaList);
-        hangul->hanjaList = NULL;
-    }
+    FcitxHangulFreeHanjaList(hangul);
 
     hic_preedit = hangul_ic_get_preedit_string (hangul->ic);
 
@@ -425,14 +519,16 @@ void FcitxHangulUpdateLookupTable(FcitxHangul* hangul)
             lookupMethod = LOOKUP_METHOD_SUFFIX;
         }
     } else {
-        FcitxInstanceGetSurroundingText(hangul->owner, FcitxInstanceGetCurrentIC(hangul->owner), &surroundingStr, &cursorPos, &anchorPos);
-        if (cursorPos != anchorPos) {
-            // If we have selection in surrounding text, we use that.
-            hanjaKey = GetSubstring(surroundingStr, cursorPos, anchorPos);
-            lookupMethod = LOOKUP_METHOD_EXACT;
-        } else {
-            hanjaKey = GetSubstring (surroundingStr, (long) cursorPos - 64, cursorPos);
-            lookupMethod = LOOKUP_METHOD_SUFFIX;
+        if (checkSurrounding) {
+            FcitxInstanceGetSurroundingText(hangul->owner, FcitxInstanceGetCurrentIC(hangul->owner), &surroundingStr, &cursorPos, &anchorPos);
+            if (cursorPos != anchorPos) {
+                // If we have selection in surrounding text, we use that.
+                hanjaKey = GetSubstring(surroundingStr, cursorPos, anchorPos);
+                lookupMethod = LOOKUP_METHOD_EXACT;
+            } else {
+                hanjaKey = GetSubstring (surroundingStr, (long) cursorPos - 64, cursorPos);
+                lookupMethod = LOOKUP_METHOD_SUFFIX;
+            }
         }
     }
 
@@ -464,7 +560,7 @@ void FcitxHangulUpdateLookupTable(FcitxHangul* hangul)
                 unsigned int* idx = fcitx_utils_malloc0(sizeof(unsigned int));
                 *idx = i;
                 word.strWord = strdup(value);
-                word.wordType = MSG_INPUT;
+                word.wordType = (i == 0) ? MSG_CANDIATE_CURSOR : MSG_INPUT;
                 word.strExtra = NULL;
                 word.extraType = MSG_INPUT;
                 word.priv = idx;
@@ -472,6 +568,8 @@ void FcitxHangulUpdateLookupTable(FcitxHangul* hangul)
                 word.callback = FcitxHangulGetCandWord;
                 FcitxCandidateWordAppend(candList, &word);
             }
+
+            FcitxCandidateWordSetFocus(candList, 0);
         }
     }
 }
@@ -480,7 +578,7 @@ void FcitxHangulFlush(FcitxHangul* hangul)
 {
     const ucschar *str;
 
-    FcitxInstanceCleanInputWindowDown(hangul->owner);
+    FcitxHangulCleanLookupTable(hangul);
 
     str = hangul_ic_flush (hangul->ic);
 
@@ -502,7 +600,9 @@ void FcitxHangulFlush(FcitxHangul* hangul)
 boolean FcitxHangulInit(void* arg)
 {
     FcitxHangul* hangul = (FcitxHangul*) arg;
+    boolean flag = true;
     FcitxInstanceSetContext(hangul->owner, CONTEXT_IM_KEYBOARD_LAYOUT, "us");
+    FcitxInstanceSetContext(hangul->owner, CONTEXT_DISABLE_AUTO_FIRST_CANDIDATE_HIGHTLIGHT, &flag);
     return true;
 }
 
@@ -519,7 +619,9 @@ INPUT_RETURN_VALUE FcitxHangulGetCandWords(void* arg)
     FcitxHangulUpdatePreedit(hangul);
 
     if (hangul->fh.hanjaMode) {
-        FcitxHangulUpdateLookupTable(hangul);
+        FcitxHangulUpdateLookupTable(hangul, false);
+    } else {
+        FcitxHangulCleanLookupTable(hangul);
     }
 
     return IRV_DISPLAY_CANDWORDS;
@@ -549,15 +651,18 @@ INPUT_RETURN_VALUE FcitxHangulGetCandWord (void* arg, FcitxCandidateWord* candWo
     if (!key || !value || !hic_preedit)
         return IRV_CLEAN;
 
+    // FcitxLog(INFO, "%s", key);
     key_len = fcitx_utf8_strlen(key);
     preedit_len = ustring_length(hangul->preedit);
     hic_preedit_len = ucs4_strlen (hic_preedit);
 
+    boolean surrounding = false;
     if (hangul->lastLookupMethod == LOOKUP_METHOD_PREFIX) {
         if (preedit_len == 0 && hic_preedit_len == 0) {
             /* remove surrounding_text */
             if (key_len > 0) {
                 FcitxInstanceDeleteSurroundingText (hangul->owner, FcitxInstanceGetCurrentIC(hangul->owner), -key_len , key_len);
+                surrounding = true;
             }
         } else {
             /* remove preedit text */
@@ -590,13 +695,21 @@ INPUT_RETURN_VALUE FcitxHangulGetCandWord (void* arg, FcitxCandidateWord* candWo
         }
 
         /* remove surrounding_text */
-        if (key_len > 0) {
+        if (LOOKUP_METHOD_EXACT != hangul->lastLookupMethod && key_len > 0) {
             FcitxInstanceDeleteSurroundingText (hangul->owner, FcitxInstanceGetCurrentIC(hangul->owner), -key_len , key_len);
+            surrounding = true;
         }
     }
 
     FcitxInstanceCommitString(hangul->owner, FcitxInstanceGetCurrentIC(hangul->owner), value);
-    return IRV_DISPLAY_CANDWORDS;
+    if (surrounding) {
+        FcitxInstanceCleanInputWindowUp(hangul->owner);
+        FcitxHangulCleanLookupTable(hangul);
+        return IRV_DISPLAY_MESSAGE;
+    }
+    else {
+        return IRV_DISPLAY_CANDWORDS;
+    }
 }
 
 /**
@@ -631,7 +744,6 @@ void* FcitxHangulCreate (FcitxInstance* instance)
 
     free(path);
 
-
     hangul->ic = hangul_ic_new(keyboard[hangul->fh.keyboardLayout]);
     hangul_ic_connect_callback (hangul->ic, "transition",
                                 FcitxHangulOnTransition, hangul);
@@ -643,7 +755,7 @@ void* FcitxHangulCreate (FcitxInstance* instance)
     iface.DoInput = FcitxHangulDoInput;
     iface.GetCandWords = FcitxHangulGetCandWords;
     iface.ReloadConfig = ReloadConfigFcitxHangul;
-    iface.UpdateSurroundingText = FcitxHangulUpdateSurroundingText;
+    iface.OnClose = FcitxHangulOnClose;
 
     FcitxInstanceRegisterIMv2(instance,
                     hangul,
@@ -676,6 +788,18 @@ void* FcitxHangulCreate (FcitxInstance* instance)
     return hangul;
 }
 
+void FcitxHangulOnClose(void* arg, FcitxIMCloseEventType event)
+{
+    FcitxHangul* hangul = arg;
+    if (event == CET_LostFocus) {
+    } else if (event == CET_ChangeByInactivate) {
+        FcitxHangulFlush(hangul);
+    } else if (event == CET_ChangeByUser) {
+        FcitxHangulFlush(hangul);
+    }
+}
+
+
 void FcitxHangulToggleHanja(void* arg)
 {
     FcitxHangul* hangul = (FcitxHangul*) arg;
@@ -692,6 +816,9 @@ void FcitxHangulUpdateHanjaStatus(FcitxHangul* hangul)
     else {
         FcitxUISetStatusString(hangul->owner, "hanja", "\xed\x95\x9c", _("Use Hangul"));
     }
+    FcitxHangulFlush(hangul);
+    FcitxHangulUpdatePreedit(hangul);
+    FcitxUIUpdateInputWindow(hangul->owner);
 }
 
 boolean FcitxHangulGetHanja(void* arg)
@@ -780,11 +907,3 @@ void FcitxHangulResetEvent(void* arg)
         FcitxUISetStatusVisable(hangul->owner, "hanja", true);
     }
 }
-
-void FcitxHangulUpdateSurroundingText(void* arg)
-{
-    FcitxHangul* hangul = (FcitxHangul*) arg;
-    FcitxInstanceCleanInputWindow(hangul->owner);
-    FcitxHangulGetCandWords(hangul);
-    FcitxUIUpdateInputWindow(hangul->owner);
-}
